#include "ParticleSystem.h"
#include "Stats.h"
D3DXMATRIX g_IdentityMatrix;

ParticleSystem::ParticleSystem(LPDIRECT3DDEVICE9 p_D3dDevice)	// singleton pattern
{
	m_Device = p_D3dDevice;

    m_VBOffset       = 0;    // Gives the offset of the vertex buffer chunk that's currently being filled
	
	HRESULT hr;
	
	if( FAILED( hr = m_Device->CreateVertexBuffer( 
		m_Discard * sizeof(ParticleVertex), 
        D3DUSAGE_DYNAMIC | D3DUSAGE_WRITEONLY | D3DUSAGE_POINTS, 
        ParticleVertex::FVF_Flags, // Our custom FVF
        D3DPOOL_DEFAULT, 
        &m_VertexBuffer, NULL )) || FAILED(hr))
    {
		Tools::LOG("ParticleSystem::ParticleSystem @ cannot CreateVertexBuffer ",true);
	}


	D3DXMatrixIdentity(&g_IdentityMatrix);

	m_PreviousRenderTime =  timeGetTime();

	int i = 0;

	for( i = 0 ; i < 1000 ; i++ )
	{
		m_DeadParticlesPool.push_back( new Particle(COLLISION_GROUP_NONE) );
	}
	for( i = 0 ; i < 100 ; i++ )
	{
		m_DeadParticleGroupsPool.push_back( new ParticleGroup(NULL,COLLISION_GROUP_NONE));
	}
	initTextures();
}


ParticleSystem::~ParticleSystem(void)
{
	delete &textureMap;

	delete &m_AliveParticleGroupsPool;

	delete &m_DeadParticleGroupsPool;

}


void	ParticleSystem::initTextures ( void )
{
	if ( FAILED (addTexture( "particle.bmp" ) ) )
	{
		Tools::LOG("ParticleSystem::initTextures @ Cannot init texture ");
	}
}


HRESULT	ParticleSystem::addTexture	( string p_FileName )
{
	LPDIRECT3DTEXTURE9	texture;
	HRESULT hr = D3DXCreateTextureFromFile( m_Device, p_FileName.c_str(), &texture );
	m_Device->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	m_Device->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);

	if (FAILED( hr ))
	{
		Tools::LOG("ParticleSystem::addTexture @ D3DXCreateTextureFromFile failed : " + p_FileName );
	}
	else
	{
		Tools::LOG("ParticleSystem::addTexture @ successfully created : " + p_FileName );
		textureMap[p_FileName] = texture;
	}
	return hr;
}

/*
* removes the particle group from the alive particles pool to dead particles pool
*/
void	ParticleSystem::removeParticleGroup	( vector<ParticleGroup*>::iterator& p_pIter )
{
	try
	{
		
		ParticleGroup *tmpPG;
		tmpPG = (*p_pIter);
		removeParticles(tmpPG->m_ParticleLayout->m_NumOfParticles,&(tmpPG->m_Particles));
		m_AliveParticleGroupsPool.erase ( p_pIter );
		m_DeadParticleGroupsPool.push_back ( tmpPG );
		PhysicsEngine::getSingleton()->unRegisterPhyObject(tmpPG);
	}
	catch (...)
	{
		Tools::LOG("ParticleSystem::removeParticleGroup @ exception : ",true);
	}
}

/*
*	get particles from the alive particles pool, if there is not
*	enough particles, creates.
*/
void	ParticleSystem::createParticles		( ParticleGroupLayout* p_Layout )
{
	try
	{
		static DWORD prevTime = 0;

		prevTime = timeGetTime();

		ParticleGroup *tmpPG;
		if( m_DeadParticleGroupsPool.size() > 0 )
		{
			tmpPG = m_DeadParticleGroupsPool[0];
			m_DeadParticleGroupsPool.erase(m_DeadParticleGroupsPool.begin());
			tmpPG->initParticles(p_Layout);
			tmpPG->m_IsDead = false;
		}
		else
		{
			tmpPG = new ParticleGroup(p_Layout,COLLISION_GROUP_NONE);
		}

		PhysicsEngine::getSingleton()->registerPhyObject(tmpPG);
		m_AliveParticleGroupsPool.push_back(tmpPG);
		Stats::getSingleton()->updateParticleTime(timeGetTime() - prevTime);
		Stats::getSingleton()->addParticles(p_Layout->m_NumOfParticles);
	}
	catch(...)
	{
		Tools::LOG("ParticleSystem::createParticles @ exception : ",true);
	}
}

void	ParticleSystem::update	( float p_ElapsedTime )
{


}

void	ParticleSystem::render	(  )
{
	try
	{
		int szOfAliveV = m_AliveParticleGroupsPool.size();
		if(szOfAliveV == 0 )
		{
			m_PreviousRenderTime = timeGetTime();
			return;
		}


		DWORD currentTime = timeGetTime();
		HRESULT hr;
		ParticleVertex *tmpPVertex;
		float elapsedTime = currentTime - m_PreviousRenderTime;

		update( elapsedTime );



		//
		// Set the render states for using point sprites..
		//

		Tools::pushRenderState(m_Device, D3DRS_ZWRITEENABLE);
		Tools::pushRenderState(m_Device, D3DRS_ZENABLE);
		Tools::pushRenderState(m_Device, D3DRS_ALPHABLENDENABLE);
		Tools::pushRenderState(m_Device, D3DRS_DESTBLEND);
		Tools::pushRenderState(m_Device, D3DRS_POINTSPRITEENABLE);
		Tools::pushRenderState(m_Device, D3DRS_POINTSCALEENABLE);
		Tools::pushRenderState(m_Device, D3DRS_POINTSIZE_MIN);
		Tools::pushRenderState(m_Device, D3DRS_POINTSCALE_A);
		Tools::pushRenderState(m_Device, D3DRS_POINTSCALE_B);
		Tools::pushRenderState(m_Device, D3DRS_POINTSCALE_C);

		m_Device->SetRenderState( D3DRS_ZWRITEENABLE, FALSE );
		m_Device->SetRenderState(D3DRS_ZENABLE,D3DZB_TRUE);

		m_Device->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
		m_Device->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_ONE );
		
		
		m_Device->SetRenderState( D3DRS_POINTSPRITEENABLE, TRUE );       // Turn on point sprites
		m_Device->SetRenderState( D3DRS_POINTSCALEENABLE,  TRUE );       // Allow sprites to be scaled with distance
		m_Device->SetRenderState( D3DRS_POINTSIZE_MIN, Tools::FtoDW(1.0f) );    // Float value that specifies the minimum size of point primitives. Point primitives are clamped to this size during rendering. 
		m_Device->SetRenderState( D3DRS_POINTSCALE_A,  Tools::FtoDW(0.0f) );    // Default 1.0
		m_Device->SetRenderState( D3DRS_POINTSCALE_B,  Tools::FtoDW(0.0f) );    // Default 0.0
		m_Device->SetRenderState( D3DRS_POINTSCALE_C,  Tools::FtoDW(1.0f) );    // Default 0.0



		// Move the offset forward so we can fill the next chunk of the vertex buffer
		m_VBOffset += m_Flush;

		// If we're about to overflow the buffer, reset the offset counter back to 0
		m_VBOffset = m_VBOffset % m_Discard;

		if( FAILED( hr = m_VertexBuffer->Lock( m_VBOffset * sizeof(ParticleVertex), // Offset to lock
									  m_Flush * sizeof(ParticleVertex),    // Size to lock
									  (void**) &tmpPVertex, 
									  m_VertexBuffer ? D3DLOCK_NOOVERWRITE : D3DLOCK_DISCARD)))
		{
			Tools::LOG("ParticleSystem::render @ cannot m_VertexBuffer->Lock - 1",true);
		}


		register int j;
		register int i;
		register int nOfParticlesInGruop	= 0;

		int nOfParticlesToRender = 0;
		
		ParticleGroup *tmpPG;
		m_Device->SetTransform(D3DTS_WORLD,&g_IdentityMatrix);	// must be identitiy matrix
																// since each particle hold its own position


		// for all the alive particle group
		for( i = 0 ; i < szOfAliveV ; i++ )
		{
			tmpPG = m_AliveParticleGroupsPool[i];
			nOfParticlesInGruop = tmpPG->m_ParticleLayout->m_NumOfParticles;

			m_Device->SetRenderState( D3DRS_POINTSIZE,     Tools::FtoDW(tmpPG->m_ParticleLayout->m_SizeOfParticle * 10.0f) ); // Float value that specifies the size to use for point size computation in cases where point size is not specified for each vertex.
			m_Device->SetTexture( 0, tmpPG->m_Texture );
			
			if( timeGetTime() - tmpPG->m_LastUpdate > 20 )	// the update time must be fixed
			{
				tmpPG->m_ParticleLayout->updateParticles(&(tmpPG->m_Particles));
				tmpPG->m_LastUpdate = timeGetTime();
			}

			//for all the particles in a group
			for ( j = 0 ; j < nOfParticlesInGruop ; j++ )
			{

				if( !tmpPG->m_Particles[j]->m_IsAlive )
					continue;


				nOfParticlesToRender++;

				tmpPVertex->m_Position	= tmpPG->m_Particles[j]->m_Position ;
				tmpPVertex->m_Color		= tmpPG->m_Particles[j]->m_Color ;

				tmpPVertex++;

				// if the current chunk is full, render all the particles in the cunk
				// and change the chunk
				if( nOfParticlesToRender == m_Flush )
				{
					m_VertexBuffer->Unlock();

					m_Device->SetStreamSource( 0, m_VertexBuffer, 0, sizeof(ParticleVertex) );
					m_Device->SetFVF( ParticleVertex::FVF_Flags );
					
					if( FAILED(hr = m_Device->DrawPrimitive( D3DPT_POINTLIST, 
						m_VBOffset, nOfParticlesToRender)))
					{
						Tools::LOG("ParticleSystem::render @ cannot m_Device->DrawPrimitive - 1",true);
					}//	if

					m_VBOffset += m_Flush;

					m_VBOffset = m_VBOffset % m_Discard;

					if( FAILED( hr = m_VertexBuffer->Lock( 
							m_VBOffset * sizeof(ParticleVertex), // Offset to lock
							m_Flush    * sizeof(ParticleVertex), // Size to lock
							(void**) &tmpPVertex, 
							m_VBOffset ? D3DLOCK_NOOVERWRITE : D3DLOCK_DISCARD)))
					{
						Tools::LOG("ParticleSystem::render @ cannot m_VertexBuffer->Lock - 2",true);
					}//	if

					nOfParticlesToRender = 0;
				}//	if
			}//	for 

			tmpPG->m_RemainingLife -= elapsedTime;
			if( tmpPG->m_RemainingLife < 0 )
			{
				this->removeParticleGroup( m_AliveParticleGroupsPool.begin() + i );
				szOfAliveV--;
			}//	if


		}//	for

		// Unlock the vertex buffer
		m_VertexBuffer->Unlock();

		// Render any remaining particles
		if( nOfParticlesToRender )
		{
			m_Device->SetStreamSource( 0, m_VertexBuffer, 0, sizeof(ParticleVertex) );
			m_Device->SetFVF( ParticleVertex::FVF_Flags );

			if(FAILED(hr = m_Device->DrawPrimitive( D3DPT_POINTLIST, m_VBOffset, 
													  nOfParticlesToRender )))
			{
				Tools::LOG("ParticleSystem::render @ cannot m_Device->DrawPrimitive - 2",true);
			}//	if
		}//	if

		//
		// Reset render states...
		//

		Tools::popRenderState(m_Device, 10);


		m_PreviousRenderTime = timeGetTime();

		Stats::getSingleton()->updateParticleTime( m_PreviousRenderTime - currentTime );

	}
	catch(...)
	{
		Tools::LOG("ParticleSystem::render @ exception on render particle system" , true);
	}

}

ParticleSystem *ms_ParticleSystemInstance = NULL;

ParticleSystem* ParticleSystem::getSingleton()
{
	if (ms_ParticleSystemInstance == NULL)
	{
		return NULL;
	}
	else
	{
		return ms_ParticleSystemInstance;
	}
}

void ParticleSystem::create( LPDIRECT3DDEVICE9 pd3dDevice )
{
	ms_ParticleSystemInstance = new ParticleSystem(pd3dDevice);
}




void	ParticleSystem::giveParticles	( unsigned int p_NumOfParticles, Particle*** p_dpParticleArray)
{
	int szOfParticlePool = m_DeadParticlesPool.size() ;
	int minNOfParticles = min( szOfParticlePool, p_NumOfParticles ) ;
	
	if ((*p_dpParticleArray) != NULL)
	{
		delete  (*p_dpParticleArray) ;
//		(*p_dpParticleArray) = NULL;
	}

	(*p_dpParticleArray) = new Particle* [p_NumOfParticles];


	register int i;
	for( i = 0  ; i < minNOfParticles ; i++ )
	{
		(*p_dpParticleArray) [i] = m_DeadParticlesPool[i];
		(*p_dpParticleArray) [i]->m_IsAlive = true;
	}

	m_DeadParticlesPool.erase( m_DeadParticlesPool.begin(),m_DeadParticlesPool.begin() + i );

	for( ; i < p_NumOfParticles ; i++ )
	{
		(*p_dpParticleArray) [i] = new Particle(COLLISION_GROUP_NONE);
		(*p_dpParticleArray) [i]->m_IsAlive = true;
	}


}

void	ParticleSystem::removeParticles	( unsigned int p_NumOfParticles, Particle*** p_dpParticleArray)
{
	try
	{
		vector<Particle*> p = m_DeadParticlesPool;
		for( int i = 0 ; i < p_NumOfParticles ; i++)
		{
			m_DeadParticlesPool.push_back( (*p_dpParticleArray)[i] );
		}
		delete (*p_dpParticleArray);
		*p_dpParticleArray = NULL;
		Stats::getSingleton()->subParticles(p_NumOfParticles);
	}
	catch (exception e)
	{
		Tools::LOG("ParticleSystem::removeParticles @ exception : ", true);
	}


}




